from django.utils import timezone
from django.db import models
from django.db.models.query import Q
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from generic_confirmation.fields import PickledObjectField


class ConfirmationManager(models.Manager):
    def confirm(self, token):
        try:
            action = self.exclude(confirmed=True).get(token=token)
        except self.model.DoesNotExist:
            return False

        if not action.is_expired():
            obj = action.resume_form_save()
            action.confirmed = True
            action.save() # FIXME: should we delete() here?
            return obj

        return False

    def pending_for(self, instance):
        ct = ContentType.objects.get_for_model(instance)
        now = timezone.now()
        return self.exclude(confirmed=True).filter(content_type=ct,
                object_pk=instance.pk).filter(
                Q(valid_until__gt=now) | Q(valid_until__isnull=True)).count()


class DeferredAction(models.Model):
    token = models.CharField(max_length=40)
    valid_until = models.DateTimeField(null=True)
    confirmed = models.BooleanField(default=False)

    form_class = models.CharField(max_length=255)
    form_input = PickledObjectField(editable=False)
    form_prefix = models.CharField(max_length=255, blank=True, null=True)

    content_type = models.ForeignKey(ContentType, null=True)
    object_pk = models.TextField(null=True)
    instance_object = GenericForeignKey('content_type', 'object_pk')

    objects = ConfirmationManager()

    def get_resume_form(self):
        form_class_name = self.form_class
        dot_index = form_class_name.rindex('.')
        module = form_class_name[:dot_index]
        klass = form_class_name[dot_index+1:]
        form_module = __import__(module, {}, {}, [''])
        form_class = getattr(form_module, klass)

        if self.instance_object is None:
            form = form_class(self.form_input)
        else:
            form = form_class(
                self.form_input, instance=self.instance_object,
                prefix=self.form_prefix)

        return form

    def resume_form_save(self, commit=True):
        form = self.get_resume_form()

        if not form.is_valid():
            raise Exception(
                "the defered form was not cleaned properly before saving")

        obj = form.save_original(commit=commit)
        if commit:
            obj.save()
        return obj

    def is_expired(self):
        if self.valid_until is None:
            return False
        now = timezone.now()
        return self.valid_until < now
    is_expired.boolean = True



from django.shortcuts import render
from django.template import RequestContext
from django.http import HttpResponseRedirect
from django.contrib import messages
from generic_confirmation.forms import ConfirmationForm


def confirm_by_form(request, template_name='confirm.html',
                    success_template_name='confirmed.html',
                    success_url=None, success_message=None,
                    form_class=ConfirmationForm):
    """
    If ``success_url`` is not None a redirect to ``success_url`` will
    be issued, once the entered confirmation code was confirmed successfully.
    Optionally, of ``success_message`` is not None, it will be set via
    Django's message system.

    If ``success_url`` is None the template ``success_template_name`` will
    be rendered once the confirmation is complete. The ``success_message``
    will then be added to the template context instead of the message_set.

    """
    if request.method == 'POST':
        form = form_class(request.POST)
        if form.is_valid():
            form.save()
            if success_url is None:
                return render(request, success_template_name,
                    {'success_message': success_message})
            else:
                if success_message is not None and request.user.is_authenticated():
                    messages.add_message(request, messages.SUCCESS, success_message)
                return HttpResponseRedirect(success_url)
    else:
        form = form_class()
    return render(request, template_name, {'form': form})



def confirm_by_get(request, token, template_name='confirm.html',
                   success_template_name="confirmed.html",
                   success_url=None, success_message=None,
                   form_class=ConfirmationForm):
    form = form_class({'token': token})
    if form.is_valid():
        form.save()
        if success_url is None:
            return render(request, success_template_name,
                    {'success_message': success_message})
        else:
            if success_message is not None and request.user.is_authenticated():
                messages.add_message(request, messages.SUCCESS, success_message)
            return HttpResponseRedirect(success_url)
    else:
        return render(request, template_name, {})



import random
from django import forms
from django.db import models
from generic_confirmation.models import DeferredAction
from generic_confirmation.main import LONG
from generic_confirmation import signals


class DeferredFormMixIn(object):
    """
    This is a MixIn class, so that you can also build deferred forms
    from already existing modified ModelForm classes.
    If you build your form from scratch you want to use the ``DeferredForm``
    class.

    """

    token_format = LONG

    def _gen_token(self, format=None, step=0):
        """
        generates a unique (in terms of DeferredAction objects) token based
        on the format tuple in the form of (alphabet, length).
        """
        if format is None:
            format = self.token_format
        chars, length = format
        token = u''.join([random.choice(chars) for i in range(length)])
        try:
            DeferredAction.objects.get(token=token)
        except DeferredAction.DoesNotExist:
            return token
        if step > 9:
            raise Exception("10 attempts to generate a unique token failed.")
        return self._gen_token(format=format, step=step+1)

    def save(self, user=None, **kwargs):
        """
        replaces the ModelForm save method with our own to defer the action
        by storing the data in the db.
        Returns a unique token which is needed to confirm the action and
        resume it.
        """
        if not self.is_valid():
            raise Exception("only call save() on a form after calling is_valid().")

        form_class_name = u"%s.%s" % (self.__class__.__module__,
                                      self.__class__.__name__)

        # we save the uncleaned data here, because form.full_clean() will
        # alter the data in cleaned_data and a second run with cleaned_data as
        # form input will fail for foreignkeys and manytomany fields.
        # additionally, storing the original input is a bit safer, because
        # this is only data which was transfered over http, so we won't
        # get any pickle errors here
        data = {'form_class':form_class_name, 'form_input':self.data,
                'token':self._gen_token(), 'form_prefix': self.prefix}
        valid_until = kwargs.pop('valid_until', None)
        if valid_until is not None:
            data.update({'valid_until': valid_until})
        defer = DeferredAction.objects.create(**data)

        if self.instance is not None:
            # this extra step makes sure that ModelForms for editing and for
            # creating objects both work.
            defer.instance_object = self.instance
            defer.save()

        # inform anyone else that confirmation is requested
        signals.confirmation_required.send(sender=self._meta.model,
                                        instance=defer, user=user)

        if hasattr(self, 'send_notification') and callable(self.send_notification):
            self.send_notification(user, instance=defer)

        return defer.token

    def save_original(self, *args, **kwargs):
        """
        triggr the original ModelForm save
        """
        return forms.ModelForm.save(self, *args, **kwargs)


class DeferredForm(DeferredFormMixIn, forms.ModelForm):
    """
    Inherit from this form to get a ModelForm which will
    automatically defer it's save method until the action
    is confirmed.
    """


class ConfirmationForm(forms.Form):
    """
    Form to use in views to confirm an action with a token.
    Makes sure the token exists and on calling ``save()``
    will resume the defered action.

    """
    token = forms.CharField(required=True)

    def clean_token(self):
        try:
            obj = DeferredAction.objects.get(token=self.cleaned_data['token'])
            return self.cleaned_data['token']
        except DeferredAction.DoesNotExist:
            raise forms.ValidationError(u"wrong token") #FIXME: i18n

    def save(self):
        return DeferredAction.objects.confirm(token=self.cleaned_data['token'])



from django.template import Library
from generic_confirmation.models import DeferredAction

register = Library()

@register.simple_tag
def pending_confirmations(instance):
    """
    {% load generic_confirmation_tags %}
    Pending: {% pending_confirmations object %}
    """
    return DeferredAction.objects.pending_for(instance)



from django.contrib import admin
from generic_confirmation.models import DeferredAction

admin.site.register(
    DeferredAction,
    list_display=('token', 'valid_until', 'confirmed', 'is_expired')
)



from django.conf.urls import url, include
from generic_confirmation import views

urlpatterns = [
    url(r'^$', views.confirm_by_form, {}, name="generic_confirmation_by_form"),
    url(r'^(?P<token>\w+)$', views.confirm_by_get, {}, name="generic_confirmation_by_get"),
]



# -*- coding: utf-8 -*-
"""Unit testing for django-generic-confirmation."""

import time
from django import VERSION
from django import forms
from django.utils import timezone
from django.test import TestCase, override_settings
from django.test.client import Client
from django.contrib.auth.models import User, Group
from django.db import models
from django.core import mail
from django.core.urlresolvers import reverse
from django.conf import settings
from django.template import Template, Context, TemplateDoesNotExist
from django.http.request import QueryDict
from generic_confirmation.fields import PickledObjectField
from generic_confirmation.forms import DeferredForm, ConfirmationForm
from generic_confirmation.models import DeferredAction
from generic_confirmation.main import LONG, SHORT, SHORT_UPPER
from generic_confirmation import signals

if VERSION < (1, 9):
    TEST_SERVER_PREFIX = "http://testserver"
else:
    TEST_SERVER_PREFIX = ""

try:
    _u = unicode
except NameError:
    _u = str

class TokenTestForm(DeferredForm):
    token_format = ('a', 1)
    class Meta:
        model = User
        fields = ('email',)

class EmailChangeForm(DeferredForm):
        class Meta:
            model = User
            fields = ('email',)

class GroupNameChangeForm(DeferredForm):
    token_format = SHORT
    class Meta:
        model = Group
        fields = ('name',)

class UserCreateForm(DeferredForm):
    token_format = SHORT_UPPER
    class Meta:
        model = User
        fields = ('username', 'email', 'password')

class GroupChangeForm(DeferredForm):
    class Meta:
        model = User
        fields = ('groups',)


class EmailChangeWithMailForm(DeferredForm):
    def send_notification(self, user=None, instance=None):
        mail.send_mail("please confirm your new address", "Please confirm %s" % instance.token,
            settings.DEFAULT_FROM_EMAIL, [self.cleaned_data['email'],])

    class Meta:
        model = User
        fields = ('email',)

class TokenGeneratorTestCase(TestCase):
    def setUp(self):
        self.user1 = User.objects.create_user('userX', 'userX@example.com', '123456')
        self.user2 = User.objects.create_user('userY', 'userY@example.com', '123456')

    def testCollision(self):
        form1 = TokenTestForm({'email': 'xxx@example.com'}, instance=self.user1)
        self.assertTrue(form1.is_valid())
        defered1 = form1.save()

        form2 = TokenTestForm({'email': 'yyy@example.com'}, instance=self.user2)
        self.assertTrue(form2.is_valid())
        # the token format only allows one possible token, so the second attempt
        # to generate one must fail because it's a not recoverable error for us
        self.assertRaises(Exception, form2.save)

class DeferFormTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('user1', 'user1@example.com', '123456')

    def testEmailChange(self):
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user)
        self.assertTrue(form.is_valid())

        defered = form.save()
        self.assertEquals(len(defered), LONG[1])
        # refetch user-object from db before checking the email,
        # because django objects don't reflect db changes done
        # elsewhere
        user_obj = User.objects.get(username=self.user.username)
        self.assertEquals(user_obj.email, 'user1@example.com')

        # ========================
        # in practice this is the boundary, where code excution will be defered.
        # ========================

        obj = DeferredAction.objects.confirm(token=defered)
        #x = obj.resume_form_save()

        # refetch user-object from db before checking the email,
        # because django objects don't reflect db changes done
        # elsewhere
        user_obj = User.objects.get(username=self.user.username)
        self.assertEquals(user_obj.email, 'xxx@example.com')


    def testUserCreation(self):
        form = UserCreateForm({'username': 'user2', 'email': 'user2@example.com', 'password': '123456'})
        self.assertTrue(form.is_valid())
        defered = form.save()
        self.assertEquals(len(defered), SHORT_UPPER[1])
        # at this point the object must not exist.
        self.assertRaises(User.DoesNotExist, User.objects.get, username='user2')

        # ========================
        # in practice this is the boundary, where code excution will be defered.
        # ========================

        obj = DeferredAction.objects.confirm(token=defered)
        #x = obj.resume_form_save()

        # refetch user-object from db before checking the email,
        # because django objects don't reflect db changes done
        # elsewhere
        user_obj = User.objects.get(username='user2')
        self.assertEquals(user_obj.email, 'user2@example.com')


    def testConfirmViaForm(self):
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user)
        self.assertTrue(form.is_valid())

        defered = form.save()
        self.assertEquals(len(defered), LONG[1])

        # refetch user-object from db before checking the email,
        # because django objects don't reflect db changes done
        # elsewhere
        user_obj = User.objects.get(username=self.user.username)
        self.assertEquals(user_obj.email, 'user1@example.com')

        # ========================
        # in practice this is the boundary, where code excution will be defered.
        # ========================

        confirm_form = ConfirmationForm({'token': defered})
        self.assertTrue(confirm_form.is_valid())
        obj = confirm_form.save()
        self.assertNotEqual(obj, False)

        # refetch user-object from db before checking the email,
        # because django objects don't reflect db changes done
        # elsewhere
        user_obj = User.objects.get(username=self.user.username)
        self.assertEquals(user_obj.email, 'xxx@example.com')


    def testConfirmBogus(self):
        result = DeferredAction.objects.confirm('some-bogus-token-1')
        self.assertEquals(result, False)

    def testConfirmBogusViaForm(self):
        confirm_form = ConfirmationForm({'token': 'some-bogus-token-2'})
        self.assertFalse(confirm_form.is_valid())


    def testCustomValidUntil(self):
        # very similar to self.testEmailChange
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user)
        self.assertTrue(form.is_valid())

        valid_date = timezone.now()+timezone.timedelta(hours=1)
        defered = form.save(valid_until=valid_date)

        obj = DeferredAction.objects.get(token=defered)

        # the token must be valid until ``valid_date``
        self.assertEquals(obj.valid_until, valid_date)


    def testConfirmExpired(self):
        # similar to self.testCustomValidUntil
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user)
        self.assertTrue(form.is_valid())

        valid_date = timezone.now() - timezone.timedelta(hours=1)
        defered = form.save(valid_until=valid_date)

        result = DeferredAction.objects.confirm(defered)
        self.assertEquals(result, False)



class ManyToManyTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('user3', 'user3@example.com', '123456')
        self.group1 = Group.objects.create(name='first_test_group')
        self.group2 = Group.objects.create(name='second_test_group')

    def testGroupChange(self):
        form = GroupChangeForm({'groups': [self.group1.pk, self.group2.pk]}, instance=self.user)
        self.assertTrue(form.is_valid())

        defered = form.save()
        self.assertEquals(len(defered), LONG[1])

        # refetch user-object from db before checking the email,
        # because django objects don't reflect db changes done
        # elsewhere
        user_obj = User.objects.get(username=self.user.username)
        self.assertEquals(list(user_obj.groups.all()), [])

        # ========================
        # in practice this is the boundary, where code excution will be defered.
        # ========================

        obj = DeferredAction.objects.confirm(token=defered)
        #x = obj.resume_form_save()

        # refetch user-object from db before checking the email,
        # because django objects don't reflect db changes done
        # elsewhere
        user_obj = User.objects.get(username=self.user.username)
        self.assertEquals(list(user_obj.groups.all()), [self.group1, self.group2])


class SignalTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('user4', 'user4@example.com', '123456')
        self.group = Group.objects.create(name='test_group_one')

    def testCatchAllListener(self):

        def dummy_listener(sender, instance, testcase=self, **kwargs):
            """ a signal receiver which does some tests """
            testcase.assertEquals(instance.__class__, DeferredAction)

        signals.confirmation_required.connect(dummy_listener)

        # just triggr a confirmation request by changing the email
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user)
        self.assertTrue(form.is_valid())
        defered = form.save()
        self.assertEquals(len(defered), LONG[1])

        form = GroupNameChangeForm({'name': 'new_name'}, instance=self.group)
        self.assertTrue(form.is_valid())
        defered = form.save()
        self.assertEquals(len(defered), SHORT[1])

    def testClassOnlyListener(self):

        def dummy_listener(sender, instance, testcase=self, **kwargs):
            """ a signal receiver which does some tests """
            testcase.assertEquals(instance.__class__, DeferredAction)
            testcase.assertEquals(sender, User)

        signals.confirmation_required.connect(dummy_listener, sender=User)

        # just triggr a confirmation request by changing the email
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user)
        self.assertTrue(form.is_valid())
        defered = form.save()
        self.assertEquals(len(defered), LONG[1])

        form = GroupNameChangeForm({'name': 'new_name'}, instance=self.group)
        self.assertTrue(form.is_valid())
        defered = form.save()
        self.assertEquals(len(defered), SHORT[1])


    def testUserPassing(self):
        def dummy_listener(sender, instance, user, testcase=self, **kwargs):
            """ a signal receiver which does some tests """
            testcase.assertEquals(instance.__class__, DeferredAction)
            testcase.assertEquals(user, self.user)

        signals.confirmation_required.connect(dummy_listener)

        # just triggr a confirmation request by changing the email
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user)
        self.assertTrue(form.is_valid())
        defered = form.save(self.user)
        self.assertEquals(len(defered), LONG[1])

    def testNoUser(self):
        def dummy_listener(sender, instance, user, testcase=self, **kwargs):
            """ a signal receiver which does some tests """
            testcase.assertEquals(instance.__class__, DeferredAction)
            testcase.assertEquals(user, None)

        signals.confirmation_required.connect(dummy_listener)

        form = GroupNameChangeForm({'name': 'new_name'}, instance=self.group)
        self.assertTrue(form.is_valid())
        defered = form.save()
        self.assertEquals(len(defered), SHORT[1])

class NotificationTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('user4', 'user4@example.com', '123456')

    def testMailNotification(self):
        form = EmailChangeWithMailForm({'email': 'new@example.com'}, instance=self.user)
        self.assertTrue(form.is_valid())
        self.assertEquals(len(mail.outbox), 0)
        token = form.save(self.user)
        self.assertEquals(len(mail.outbox), 1)
        # make sure the right token is in the body of the message
        self.assertTrue(token in mail.outbox[0].body)


class TemplatetagTestCase(TestCase):
    def setUp(self):
        self.user5 = User.objects.create_user('user5', 'user5@example.com', '123456')
        self.user6 = User.objects.create_user('user6', 'user6@example.com', '123456')
        self.user7 = User.objects.create_user('user7', 'user7@example.com', '123456')


    def testHasPendingTokens(self):
        # generate a Token
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user5)
        self.assertTrue(form.is_valid())
        defered = form.save()
        #self.assertEquals(DeferredAction.objects.pending_for(self.user5), 1)

        t = Template("""{% load generic_confirmation_tags %}{% pending_confirmations object %}""")
        html = t.render(Context({'object': self.user5}))
        self.assertEquals(html, "1")


    def testHasNoPendingTokens(self):
        t = Template("""{% load generic_confirmation_tags %}{% pending_confirmations object %}""")
        html = t.render(Context({'object': self.user6}))
        self.assertEquals(html, "0")

    def testHasAlreadyExpiredToken(self):
        # generate a Token
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user7)
        self.assertTrue(form.is_valid())
        valid_date = timezone.now() - timezone.timedelta(hours=1)
        defered = form.save(valid_until=valid_date)
        #self.assertEquals(DeferredAction.objects.pending_for(self.user7), 0)

        t = Template("""{% load generic_confirmation_tags %}{% pending_confirmations object %}""")
        html = t.render(Context({'object': self.user7}))
        self.assertEquals(html, "0")


@override_settings(ROOT_URLCONF="generic_confirmation.tests.urls")
class ViewTestCase(TestCase):
    """
    without bundled templates this does not make too much sense, but
    at least we can have a better test-coverage buy running the parts and
    catching the proper exceptions.

    """
    def setUp(self):
        self.client = Client()
        self.user8 = User.objects.create_user('user8', 'user8@example.com', '123456')
        self.user9 = User.objects.create_user('user9', 'user9@example.com', '123456')

    def testBogusConfirmByGet(self):
        # currently there is no bundled template
        # should be fixed in a future version
        self.assertRaises(TemplateDoesNotExist, self.client.get ,reverse('generic_confirmation_by_get', kwargs={'token': 'somebogustoken3'}))

    def testValidConfirmByGet(self):
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user8)
        self.assertTrue(form.is_valid())
        defered = form.save()
        # currently there is no bundled template
        # should be fixed in a future version
        self.assertRaises(TemplateDoesNotExist, self.client.get ,reverse('generic_confirmation_by_get', kwargs={'token': defered}))

    def testValidConfirmByGetWithCustomSuccessMessage(self):
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user8)
        self.assertTrue(form.is_valid())
        defered = form.save()
        # currently there is no bundled template
        # should be fixed in a future version
        self.assertRaises(TemplateDoesNotExist, self.client.get ,reverse('generic_confirmation_by_get_with_message', kwargs={'token': defered}))

    def testValidConfirmByGetWithCustomSuccessUrl(self):
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user8)
        self.assertTrue(form.is_valid())
        defered = form.save()
        # currently there is no bundled template
        # should be fixed in a future version
        response = self.client.get(reverse('generic_confirmation_by_get_with_url', kwargs={'token': defered}))
        self.assertEquals(response.status_code, 302)
        self.assertEquals(response['Location'], TEST_SERVER_PREFIX + '/success/')

    def testValidConfirmByGetWithCustomSuccessUrlAndMessage(self):
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user8)
        self.assertTrue(form.is_valid())
        defered = form.save()
        # currently there is no bundled template
        # should be fixed in a future version
        response = self.client.get(reverse('generic_confirmation_by_get_with_url_and_message', kwargs={'token': defered}))
        self.assertEquals(response.status_code, 302)
        self.assertEquals(response['Location'], TEST_SERVER_PREFIX + '/success/')


    def testConfirmByFormGET(self):
        # currently there is no bundled template
        # should be fixed in a future version
        self.assertRaises(TemplateDoesNotExist, self.client.get ,reverse('generic_confirmation_by_form'))

    def testBogusConfirmByFormPOST(self):
        # currently there is no bundled template
        # should be fixed in a future version
        self.assertRaises(TemplateDoesNotExist, self.client.post ,reverse('generic_confirmation_by_form'), {'token': 'some-bogus-token-4',})

    def testValidConfirmByFormPOST(self):
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user9)
        self.assertTrue(form.is_valid())
        defered = form.save()
        # currently there is no bundled template
        # should be fixed in a future version
        self.assertRaises(TemplateDoesNotExist, self.client.post ,reverse('generic_confirmation_by_form'), {'token': defered,})

    def testValidConfirmByFormPOSTWithCustomSuccessMessage(self):
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user9)
        self.assertTrue(form.is_valid())
        defered = form.save()
        # currently there is no bundled template
        # should be fixed in a future version
        self.assertRaises(TemplateDoesNotExist, self.client.post ,reverse('generic_confirmation_by_form_with_message'), {'token': defered})

    def testValidConfirmByFormPOSTWithCustomSuccessUrl(self):
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user9)
        self.assertTrue(form.is_valid())
        defered = form.save()
        # currently there is no bundled template
        # should be fixed in a future version
        response = self.client.post(reverse('generic_confirmation_by_form_with_url'), {'token': defered})
        self.assertEquals(response.status_code, 302)
        self.assertEquals(response['Location'], TEST_SERVER_PREFIX + '/success/')

    def testValidConfirmByFormPOSTWithCustomSuccessUrlAndMessage(self):
        form = EmailChangeForm({'email': 'xxx@example.com'}, instance=self.user9)
        self.assertTrue(form.is_valid())
        defered = form.save()
        # currently there is no bundled template
        # should be fixed in a future version
        response = self.client.post(reverse('generic_confirmation_by_form_with_url_and_message'), {'token': defered})
        self.assertEquals(response.status_code, 302)
        self.assertEquals(response['Location'], TEST_SERVER_PREFIX + '/success/')


# taken (but modified) from djangosnippets.org/snippets/513 by obeattie

class TestingModel(models.Model):
	pickle_field = PickledObjectField()

class TestCustomDataType(str):
	pass

class PickledObjectFieldTests(TestCase):
    def setUp(self):
        self.testing_data = (
            {1:1, 2:4, 3:6, 4:8, 5:10},
            'Hello World',
            (1, 2, 3, 4, 5),
            [1, 2, 3, 4, 5],
            TestCustomDataType('Hello World'),
            _u(u"\xf3"),  # regression test for non-latin1 encodings in pickled data
            QueryDict("email=test@example.com&test=1&test=2"),  # mostly used datatype
        )
        return super(PickledObjectFieldTests, self).setUp()

    def testDataIntegrity(self):
        """Tests that data remains the same when saved to and fetched from the database."""
        for value in self.testing_data:
            model_test = TestingModel(pickle_field=value)
            model_test.save()
            model_test = TestingModel.objects.get(id__exact=model_test.id)
            self.assertEquals(value, model_test.pickle_field)
            model_test.delete()

    def testExactLookups(self):
        """Tests that lookups can be performed on data once stored in the database."""
        for value in self.testing_data:
            model_test = TestingModel(pickle_field=value)
            model_test.save()
            self.assertEquals(value, TestingModel.objects.get(pickle_field__exact=value).pickle_field)
            model_test.delete()

    def testInLookups(self):
        """Tests that lookups can be performed on data once stored in the database."""
        for value in self.testing_data:
            model_test = TestingModel(pickle_field=value)
            model_test.save()
            self.assertEquals(value, TestingModel.objects.filter(pickle_field__in=[value,])[0].pickle_field)
            model_test.delete()

    def testBogusLookup(self):
        """Tests that bogus lookups raise an TypeError."""
        for value in self.testing_data:
            model_test = TestingModel(pickle_field=value)
            model_test.save()
            self.assertRaises(TypeError, TestingModel.objects.filter, pickle_field__contains=[value,])
            model_test.delete()


class TestingModelForm(DeferredForm):
    class Meta:
        model = TestingModel
        exclude = ()


class FormPrefixTests(TestCase):
    def testFormPrefix(self):
        """
        Testing the situation when deferred form has prefix. Prefixes also
        should be saved, otherwise form_input will not be accepted when
        save operation will be resumed.

        """
        PREFIX = 'test_prefix'

        model = TestingModel.objects.create(pickle_field='none')

        model_form = TestingModelForm(instance=model, prefix=PREFIX)

        data = {f.html_name: str(f.value() + '-changed') for f in model_form}

        unprefixed_form = TestingModelForm(data, instance=model)
        self.assertFalse(unprefixed_form.is_valid())

        model_form = TestingModelForm(data, instance=model, prefix=PREFIX)
        self.assertTrue(model_form.is_valid())
        self.assertTrue(model_form.has_changed())

        token = model_form.save()

        form = ConfirmationForm({'token': token})
        self.assertTrue(form.is_valid())
        action = DeferredAction.objects.get(token=token)

        resume_form = action.get_resume_form()

        self.assertEqual(
            resume_form.prefix, model_form.prefix,
            msg="Resume form has to have the same prefix as initial form")

        model = TestingModel.objects.get(pk=model.pk)
        self.assertEqual(model.pickle_field, 'none')
        form.save()
        model = TestingModel.objects.get(pk=model.pk)
        self.assertEqual(model.pickle_field, 'none-changed')



from django.conf.urls import url
from generic_confirmation import views

urlpatterns = [
    url(r'^by-form/$', views.confirm_by_form,
        {},
        name="generic_confirmation_by_form"),

    url(r'^by-form-with-message/$', views.confirm_by_form,
        {'success_message': "This is a success message"},
        name="generic_confirmation_by_form_with_message"),

    url(r'^by-form-with-url/$', views.confirm_by_form,
        {'success_url': "/success/"},
        name="generic_confirmation_by_form_with_url"),

    url(r'^by-form-with-url-and-message/$', views.confirm_by_form,
        {'success_url': "/success/",
         'success_message': "This is a success message"},
        name="generic_confirmation_by_form_with_url_and_message"),



    url(r'^by-get/(?P<token>\w+)$', views.confirm_by_get,
        {},
        name="generic_confirmation_by_get"),

    url(r'^by-get-with-message/(?P<token>\w+)$', views.confirm_by_get,
        {'success_message': "This is a success message"},
        name="generic_confirmation_by_get_with_message"),

    url(r'^by-get-with-url/(?P<token>\w+)$', views.confirm_by_get,
        {'success_url': "/success/"},
        name="generic_confirmation_by_get_with_url"),

    url(r'^by-get-with-url-and-message/(?P<token>\w+)$', views.confirm_by_get,
        {'success_url': "/success/",
         'success_message': "This is a success message"},
        name="generic_confirmation_by_get_with_url_and_message"),
]


# token format definitions
# (<alphabet>, <length>)
# alphabet here is without ijl10O which could easily be misread in some fonts
LONG = ('abcdefghkmnopqrstuvwwxyzABCDEFGHKLMNPQRSTUVWXYZ23456789', 24) # for emails
SHORT = ('abcdefghkmnopqrstuvwwxyzABCDEFGHKLMNPQRSTUVWXYZ23456789', 6) # because we can =)
SHORT_UPPER = ('ABCDEFGHKLMNPQRSTUVWXYZ23456789', 6) # for sms



from django.dispatch import Signal

# sender is the class which is edited, instance is the DeferedAction instance
# user is the user who edited the sender instance or None. user is only not
# None if passed to form form.save(user=request.user) method ... 
confirmation_required = Signal(providing_args=["instance", "user"])



import os.path
from django import VERSION

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'generic_confirmation',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

SITE_ID = 1
SECRET_KEY = '+zzix-&k$afk-k0d0s7v01w0&15z#ne$71qf28#e$$c*@g742z'

ROOT_URLCONF = "urls"

DEBUG = False

LANGUAGE = "en-us"

STATIC_URL = '/static/'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(os.path.dirname(__file__), 'database.sqlite'),
    }
}

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]


# prevent migrations from running during tests, because test models have none
if VERSION < (1,9):
    MIGRATION_MODULES = {
        'auth': 'tests.migrations.auth',
        'contenttypes': 'tests.migrations.contenttypes',
        'admin': 'tests.migrations.admin',
        'sessions': 'tests.migrations.sessions',
        'generic_confirmation': 'tests.migrations.generic_confirmation',
    }
else:
    MIGRATION_MODULES = {
        'auth': None,
        'contenttypes': None,
        'admin': None,
        'sessions': None,
        'generic_confirmation': None,
    }



import base64
# based on djangosnippets.org/snippets/513 by obeattie
from django.db import models
from django.utils.encoding import smart_bytes

try:
    import cPickle as pickle
except ImportError:
    import pickle


class PickledObject(bytes):
    """
    A subclass of bytes (str in python2) so it can be told whether a string is
    a pickled object or not (if the object is an instance of this class then
    it must [well, should] be a pickled one).

    """
    pass


class PickledObjectField(models.Field):

    def from_db_value(self, value, expression, connection, context):
        if isinstance(value, PickledObject):
            # If the value is a definite pickle; and an error is raised in de-pickling
            # it should be allowed to propogate.
            return pickle.loads(smart_bytes(base64.b64decode(value)))
        else:
            try:
                return pickle.loads(smart_bytes(base64.b64decode(value)))
            except:
                # If an error was raised, just return the plain value
                return value

    def get_prep_value(self, value):
        if value is not None and not isinstance(value, PickledObject):
            value = base64.b64encode(PickledObject(pickle.dumps(value))).decode()
        return value

    def get_internal_type(self):
        return 'TextField'

    def get_lookup(self, lookup_name):
        if lookup_name in ('exact', 'in'):
            return super(PickledObjectField, self).get_lookup(lookup_name)
        else:
            raise TypeError('Lookup type %s is not supported.' % lookup_name)



#!/usr/bin/env python
import os
import sys
from django import VERSION

sys.path.insert(0, os.path.join(os.path.dirname(__file__), os.path.pardir))

if __name__ == "__main__":
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")

    if VERSION < (1, 6):
        test_runners_args = {
            'TEST_RUNNER': 'discover_runner.DiscoverRunner',
        }
    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)



from django.conf.urls import url, include
from django.contrib import admin

urlpatterns = [
    url(r'^admin/', include(admin.site.urls)),
    url(r'^confirm/', include('generic_confirmation.urls')),
]



from distutils.core import setup

setup(
    name='django-generic-confirmation',
    version=__import__('generic_confirmation').__version__,
    description='Deferred forms for the Django webframework',
    author='Arne Brodowski',
    author_email='mail@arnebrodowski.de',
    url='https://github.com/arneb/django-generic-confirmation/',
    packages=(
        'generic_confirmation',
        'generic_confirmation.templatetags',
        'generic_confirmation.tests',
    ),
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Utilities',
        'Framework :: Django',
    ),
)
